// $Id: Options.cpp 865 2015-04-21 22:41:49Z ge $
/// \file Options.h
/// \brief contains the Options class
///
/// $Revision: 865 $
/// \author Gerald Weber <gweberbh@gmail.com>

#include "Options.h"
#include <string>
#include <deque>
#include <iomanip>
#include <sstream>
#include <iterator>
#include <cstdlib>

namespace gbc
{
template<class _Tp>
inline std::ostream& operator<<(std::ostream& out,const std::deque<_Tp>& dq)
  {
  std::copy(dq.begin(),dq.end(),std::ostream_iterator<typename std::deque<_Tp>::value_type>(out,","));
  return out;
  }

template<class _Tp1,class _Tp2>
inline std::ostream& operator<<(std::ostream& out,const std::map<_Tp1,_Tp2>& mp)
  {
  typename std::map<_Tp1,_Tp2>::const_iterator it;
  for (it=mp.begin(); it != mp.end(); ++it)
    out << it->first << ":" << it->second << ",";
  return out;
  }

size_t OptionMap::Max_arg_length=0;
size_t OptionMap::Max_desc_length=0;
size_t OptionMap::Max_type_desc_length=0;
bool   OptionMap::Debug=false;

OptionMap::map_type OptionMap::all_options=OptionMap::map_type(); 

OptionMap::arg_deque_type OptionMap::all_arguments=OptionMap::arg_deque_type();

OptionMap::arg_map_type OptionMap::interpreted_arguments=OptionMap::arg_map_type();

void OptionMap::add_option(std::string arg, BaseOption* bo)
  {
  if (all_options.find(arg) == all_options.end())
    {
    all_options.insert(map_type::value_type(arg,bo));
    if (arg.length() > OptionMap::Max_arg_length) OptionMap::Max_arg_length=arg.length();
    if (bo->Description.length() > OptionMap::Max_desc_length) OptionMap::Max_desc_length=bo->Description.length();
    }
  else 
    {
    std::cerr << "Option with name " << arg <<  " already defined, ending program" << std::endl;
    exit(1);
    }
  }

void OptionMap::arguments(int argc, char *argv[])
    {
    int i;
    std::string arg;
    for (i=1; i < argc; i++)
      {
      arg=argv[i]; 
      all_arguments.push_back(arg);
      interpretation(arg);
      if (Debug)
        {
        std::cerr << __FILE__ << ":" << __LINE__ << " i=" << i << " arg=" << arg << std::endl;
        }
      }
    } 


void OptionMap::interpretation(std::string arg)
  {
  bool interpreted=false;
  std::string opt, argument;
  std::string::size_type equal=arg.find("=");
  std::string::size_type dash =arg.find("-");
  std::string::size_type dash2=arg.find("--");
  if (dash != std::string::npos) // if we find a dash
    {
    if (equal != std::string::npos) // if we also find an equal sign
      {
      opt=arg.substr(dash+1,equal-1); // everything between the dash and the equal sign must be the name of the command
      argument=arg.substr(equal+1);   // what comes after the equal sign must the its argument
      }
    else // if we don't find an equal sign
      {
      opt=arg.substr(dash+1); // everything after the dash is the command name
      argument=std::string(); // leave argument empty
      }
    interpreted=true;
    }
  // insert interpreted argumento into map.
  if (interpreted) 
    interpreted_arguments.insert(OptionMap::arg_map_type::value_type(opt,argument));
  }

Option<bool>   Verbose("v","verbose (v=1 shows all options, v=0 shows only provided options)",OptionMap::optional,false);

void OptionMap::finished(void)
  {
  scan_arguments();
  check_if_all_given();
  show_options();
  }

void OptionMap::check_if_all_given(void)
  {
  if (interpreted_arguments.empty())
    {
    show_options(1);
    std::cerr << "No arguments provided, showing list of available arguments and exiting" << std::endl;
    exit(1);
    }
  OptionMap::map_type::iterator            opt_it;
  for (opt_it=all_options.begin(); opt_it != all_options.end(); opt_it++)
    {
    if (!opt_it->second->Provided && opt_it->second->Strictly_needed)
       {
       show_options(1);
       std::cerr << "Necessary arguments missing, check list above for occurences of string **MISSING**" << std::endl;
       exit(1);
       }
    if (!opt_it->second->Provided && opt_it->second->Strictly_needed==-1)
       {
       show_options(1);
       std::cerr << "Two or more arguments given which can not be used together, check documentation" << std::endl;
       exit(1);
       }
    }
  }


// Set all type identification numbers
template<> int OptionMap::type_string<double>(void)           {return 1;}
template<> int OptionMap::type_string<int>(void)              {return 2;}
template<> int OptionMap::type_string<std::string>(void)      {return 3;}
template<> int OptionMap::type_string<bool>(void)             {return 4;}
template<> int OptionMap::type_string<Range<double> >(void)   {return 11;}
template<> int OptionMap::type_string<Range<int> >(void)      {return 12;}
template<> int OptionMap::type_string<std::deque<double> >(void) {return 21;}
template<> int OptionMap::type_string<std::deque<int> >(void) {return 22;}
template<> int OptionMap::type_string<std::deque<std::string> >(void) 
                                                              {return 23;}
template<> int OptionMap::type_string<std::map<std::string,double> >(void) 
                                                              {return 331;}
template<> int OptionMap::type_string<std::map<std::string,std::string> >(void) 
                                                              {return 333;}

// Brief description about the option types
template<> std::string OptionMap::type_description<double>(void)           {return "double";}
template<> std::string OptionMap::type_description<int>(void)              {return "integer";}
template<> std::string OptionMap::type_description<std::string>(void)      {return "string";}
template<> std::string OptionMap::type_description<bool>(void)             {return "boolean (0 or 1)";}
template<> std::string OptionMap::type_description<Range<double> >(void)   {return "Range (e.g. 0.1:5.2/10)";}
template<> std::string OptionMap::type_description<Range<int> >(void)      {return "Range (e.g. 0:10/2 integers only)";}
template<> std::string OptionMap::type_description<std::deque<double> >(void) {return "List of floating point numbers";}
template<> std::string OptionMap::type_description<std::deque<int> >(void) {return "List of integers";}
template<> std::string OptionMap::type_description<std::deque<std::string> >(void) {return "List of strings";}
template<> std::string OptionMap::type_description<std::map<std::string,double> >(void) 
                                                                           {return "Map of strings->double";}
template<> std::string OptionMap::type_description<std::map<std::string,std::string> >(void) 
                                                                           {return "Map of strings->strings";}
template<class _Tp>
inline void show(std::ostream& out, void* op)
  {static_cast<Option<_Tp>* >(op)->show(std::cout);}

void OptionMap::show_options(int action)
  {
  OptionMap::map_type::iterator            opt_it;
  for (opt_it=all_options.begin(); opt_it != all_options.end(); opt_it++)
    {
    if ((opt_it->second->Provided || Verbose) || action)
    {
    switch (opt_it->second->Option_type)
      {
      case 1: show<double>(std::cout,opt_it->second);
              break;
      case 2: show<int>(std::cout,opt_it->second);
              break;
      case 3: show<std::string>(std::cout,opt_it->second);
              break;
      case 4: show<bool>(std::cout,opt_it->second);
              break;
      case 11:show<Range<double> >(std::cout,opt_it->second);
              break;
      case 12:show<Range<int> >(std::cout,opt_it->second);
              break;
      case 21:show<std::deque<double> >(std::cout,opt_it->second);
              break;
      case 22:show<std::deque<int> >(std::cout,opt_it->second);
              break;
      case 23:show<std::deque<std::string> >(std::cout,opt_it->second);
              break;
      case 331:show<std::map<std::string,double> >(std::cout,opt_it->second);
              break;
      case 333:show<std::map<std::string,std::string> >(std::cout,opt_it->second);
              break;
      }
      std::cout << std::endl;
    }}
  }

bool OptionMap::provided(std::string arg)
  {
  OptionMap::map_type::iterator            opt_it;
  opt_it=all_options.find(arg);
  if (opt_it != all_options.end()) return opt_it->second->Provided;
  else return false;
  }

template<class _Tp>
inline void interp(std::stringstream& st,  void* op)
  {interp(st,static_cast<Option<_Tp>* >(op));}

template<class _Tp>
inline void interp(std::stringstream& st, Option<_Tp>* op)
  {st >> op->Value;}

template<class _Tp>
inline void interp(std::stringstream& st,  Option<Range<_Tp> >* rg)
  {
  typedef _Tp internal_type;
  char skip_char;
  internal_type val;
  typename Range<_Tp>::steps_type stp;
  st >> val >> skip_char; rg->Value.begin(val);
  if (!st.eof() && skip_char == ':') 
    {
    st >> val >> skip_char; rg->Value.end(val);
    if (!st.eof()) 
      {
      if (skip_char == ':' ) {st >> val; rg->Value.increment(val);}
      if (skip_char == '/' ) {st >> stp; rg->Value.steps(stp);}
      }
    else rg->Value.increment(internal_type(1));
    }
  else 
    {
    rg->Value.end(val);
    rg->Value.increment(internal_type()); 
    }
  rg->Value.reset();
  }

template<class _Tp>
inline void interp(std::stringstream& st, Option<std::deque<_Tp> >* dq)
  {
  std::string arg;
  st >> arg;
  std::string::size_type comma_i=0,comma_f=arg.find(',');
  if (comma_f == std::string::npos) comma_f=arg.size();
  _Tp value;
  while (comma_i < comma_f)
    {
    std::string field=arg.substr(comma_i,comma_f-comma_i);
    std::stringstream value_field;
    value_field << field;

    value_field >> value;
    dq->Value.push_back(value);

    comma_i=comma_f+1;
    comma_f=arg.find(',',comma_i);
    if (comma_f == std::string::npos) comma_f=arg.size();
    }
  }

template<class _Tp1,class _Tp2>
inline void interp(std::stringstream& st, Option<std::map<_Tp1,_Tp2> >* mp)
  {
  std::string arg;
  st >> arg;
  std::string::size_type comma_i=0,comma_f=arg.find(',');
  if (comma_f == std::string::npos) comma_f=arg.size();
  _Tp1 key; _Tp2 value;
  while (comma_i < comma_f)
    {
    std::string field=arg.substr(comma_i,comma_f-comma_i);
    std::string::size_type pos=field.find(':');
    std::stringstream key_field,value_field;
    key_field   << field.substr(0,pos);
    value_field << field.substr(pos+1);

    key_field >> key;
    value_field >> value;
    mp->Value[key]=value;

    comma_i=comma_f+1;
    comma_f=arg.find(',',comma_i);
    if (comma_f == std::string::npos) comma_f=arg.size();
    }
  }


void OptionMap::scan_arguments(void)
  {
  OptionMap::map_type::iterator            opt_it;
  OptionMap::arg_map_type::const_iterator  arg_it;

  // We iterate over all defined options and check it they were provided
  for (opt_it=all_options.begin(); opt_it != all_options.end(); opt_it++)
    {
    // We try to find if the option was provided and stop at first match.
    arg_it=interpreted_arguments.find(opt_it->first);
    if (arg_it != interpreted_arguments.end())
      {
      std::stringstream st(arg_it->second); // assign the argument, usually what came after the equal sign
      Option<bool>* so_bool;

      switch (opt_it->second->Option_type)
        {
        case 1: interp<double>(st,opt_it->second);
                break;

        case 2: interp<int>(st,opt_it->second);
                break;

        case 3: interp<std::string>(st,opt_it->second);
                break;

        case 4: so_bool=static_cast<Option<bool>* >(opt_it->second);
                if (arg_it->second != std::string()) st >> so_bool->Value;
                else so_bool->Value=true;
                break;

        case 11: interp<Range<double> >(st,opt_it->second);
                 break;

        case 12: interp<Range<int> >(st,opt_it->second);
                 break;

        case 21: interp<std::deque<double> >(st,opt_it->second);
                 break;

        case 22: interp<std::deque<int> >(st,opt_it->second);
                 break;

        case 23: interp<std::deque<std::string> >(st,opt_it->second);
                 break;

        case 331: interp<std::map<std::string,double> >(st,opt_it->second);
                  break;

        case 333: interp<std::map<std::string,std::string> >(st,opt_it->second);
                  break;
        }
      opt_it->second->Provided=true;
      }
    }
  }

};
